@@ -8,18 +8,18 @@ pub use self::errors::PostgresSourceError;
 pub use connection::rewrite_tls_args;
 pub use typesystem::PostgresTypeSystem;
 
+use crate::constants::DB_BUFFER_SIZE;
 use crate::{
     data_order::DataOrder,
     errors::ConnectorXError,
     sources::{PartitionParser, Produce, Source, SourcePartition},
-    sql::{count_query, get_limit, limit1_query, CXQuery},
+    sql::{count_query, CXQuery},
 };
 use anyhow::anyhow;
 use chrono::{DateTime, NaiveDate, NaiveDateTime, NaiveTime, Utc};
 use csv::{ReaderBuilder, StringRecord, StringRecordsIntoIter};
 use fehler::{throw, throws};
 use hex::decode;
-use log::debug;
 use postgres::{
     binary_copy::{BinaryCopyOutIter, BinaryCopyOutRow},
     fallible_iterator::FallibleIterator,
@@ -31,7 +31,7 @@ use r2d2_postgres::PostgresConnectionManager;
 use rust_decimal::Decimal;
 use serde_json::{from_str, Value};
 use sqlparser::dialect::PostgreSqlDialect;
-use std::io::BufRead;
+use std::convert::TryFrom;
 use std::marker::PhantomData;
 use uuid::Uuid;
 
@@ -47,6 +47,34 @@ pub enum CursorProtocol {}
 type PgManager<C> = PostgresConnectionManager<C>;
 type PgConn<C> = PooledConnection<PgManager<C>>;
 
+// take a row and unwrap the interior field from column 0
+fn convert_row<'b, R: TryFrom<usize> + postgres::types::FromSql<'b> + Clone>(row: &'b Row) -> R {
+    let nrows: Option<R> = row.get(0);
+    nrows.expect("Could not parse int result from count_query")
+}
+
+#[throws(PostgresSourceError)]
+fn get_total_rows<C>(conn: &mut PgConn<C>, query: &CXQuery<String>) -> usize
+where
+    C: MakeTlsConnect<Socket> + Clone + 'static + Sync + Send,
+    C::TlsConnect: Send,
+    C::Stream: Send,
+    <C::TlsConnect as TlsConnect<Socket>>::Future: Send,
+{
+    let dialect = PostgreSqlDialect {};
+
+    let row = conn.query_one(count_query(query, &dialect)?.as_str(), &[])?;
+    let col_type = PostgresTypeSystem::from(row.columns()[0].type_());
+    match col_type {
+        PostgresTypeSystem::Int2(_) => convert_row::<i16>(&row) as usize,
+        PostgresTypeSystem::Int4(_) => convert_row::<i32>(&row) as usize,
+        PostgresTypeSystem::Int8(_) => convert_row::<i64>(&row) as usize,
+        _ => throw!(anyhow!(
+            "The result of the count query was not an int, aborting."
+        )),
+    }
+}
+
 pub struct PostgresSource<P, C>
 where
     C: MakeTlsConnect<Socket> + Clone + 'static + Sync + Send,
@@ -55,10 +83,10 @@ where
     <C::TlsConnect as TlsConnect<Socket>>::Future: Send,
 {
     pool: Pool<PgManager<C>>,
+    origin_query: Option<String>,
     queries: Vec<CXQuery<String>>,
     names: Vec<String>,
     schema: Vec<PostgresTypeSystem>,
-    buf_size: usize,
     _protocol: PhantomData<P>,
 }
 
@@ -76,17 +104,13 @@ where
 
         Self {
             pool,
+            origin_query: None,
             queries: vec![],
             names: vec![],
             schema: vec![],
-            buf_size: 32,
             _protocol: PhantomData,
         }
     }
-
-    pub fn buf_size(&mut self, buf_size: usize) {
-        self.buf_size = buf_size;
-    }
 }
 
 impl<P, C> Source for PostgresSource<P, C>
@@ -115,55 +139,45 @@ where
         self.queries = queries.iter().map(|q| q.map(Q::to_string)).collect();
     }
 
+    fn set_origin_query(&mut self, query: Option<String>) {
+        self.origin_query = query;
+    }
+
     #[throws(PostgresSourceError)]
     fn fetch_metadata(&mut self) {
         assert!(!self.queries.is_empty());
 
         let mut conn = self.pool.get()?;
+        let first_query = &self.queries[0];
 
-        for (i, query) in self.queries.iter().enumerate() {
-            // assuming all the partition queries yield same schema
-            match conn.query_opt(limit1_query(query, &PostgreSqlDialect {})?.as_str(), &[]) {
-                Ok(Some(row)) => {
-                    let (names, types) = row
-                        .columns()
-                        .iter()
-                        .map(|col| {
-                            (
-                                col.name().to_string(),
-                                PostgresTypeSystem::from(col.type_()),
-                            )
-                        })
-                        .unzip();
+        let stmt = conn.prepare(first_query.as_str())?;
 
-                    self.names = names;
-                    self.schema = types;
+        let (names, types) = stmt
+            .columns()
+            .iter()
+            .map(|col| {
+                (
+                    col.name().to_string(),
+                    PostgresTypeSystem::from(col.type_()),
+                )
+            })
+            .unzip();
 
-                    return;
-                }
-                Ok(None) => {}
-                Err(e) if i == self.queries.len() - 1 => {
-                    // tried the last query but still get an error
-                    debug!("cannot get metadata for '{}': {}", query, e);
-                    throw!(e);
-                }
-                Err(_) => {}
+        self.names = names;
+        self.schema = types;
+    }
+
+    #[throws(PostgresSourceError)]
+    fn result_rows(&mut self) -> Option<usize> {
+        match &self.origin_query {
+            Some(q) => {
+                let cxq = CXQuery::Naked(q.clone());
+                let mut conn = self.pool.get()?;
+                let nrows = get_total_rows(&mut conn, &cxq)?;
+                Some(nrows)
             }
+            None => None,
         }
-
-        // tried all queries but all get empty result set
-
-        // try to use COPY command get the column headers
-        let copy_query = format!("COPY ({}) TO STDOUT WITH CSV HEADER", self.queries[0]);
-        let mut reader = conn.copy_out(&*copy_query)?;
-        let mut buf = String::new();
-        reader.read_line(&mut buf)?;
-        self.names = buf[0..buf.len() - 1] // remove last '\n'
-            .split(',')
-            .map(|s| s.to_string())
-            .collect();
-        // set all columns as string (align with pandas)
-        self.schema = vec![PostgresTypeSystem::Text(false); self.names.len()];
     }
 
     fn names(&self) -> Vec<String> {
@@ -184,7 +198,6 @@ where
                 conn,
                 &query,
                 &self.schema,
-                self.buf_size,
             ));
         }
         ret
@@ -203,7 +216,6 @@ where
     schema: Vec<PostgresTypeSystem>,
     nrows: usize,
     ncols: usize,
-    buf_size: usize,
     _protocol: PhantomData<P>,
 }
 
@@ -214,19 +226,13 @@ where
     C::Stream: Send,
     <C::TlsConnect as TlsConnect<Socket>>::Future: Send,
 {
-    pub fn new(
-        conn: PgConn<C>,
-        query: &CXQuery<String>,
-        schema: &[PostgresTypeSystem],
-        buf_size: usize,
-    ) -> Self {
+    pub fn new(conn: PgConn<C>, query: &CXQuery<String>, schema: &[PostgresTypeSystem]) -> Self {
         Self {
             conn,
             query: query.clone(),
             schema: schema.to_vec(),
             nrows: 0,
             ncols: schema.len(),
-            buf_size,
             _protocol: PhantomData,
         }
     }
@@ -244,17 +250,8 @@ where
     type Error = PostgresSourceError;
 
     #[throws(PostgresSourceError)]
-    fn prepare(&mut self) -> () {
-        let dialect = PostgreSqlDialect {};
-        self.nrows = match get_limit(&self.query, &dialect)? {
-            None => {
-                let row = self
-                    .conn
-                    .query_one(count_query(&self.query, &dialect)?.as_str(), &[])?;
-                row.get::<_, i64>(0) as usize
-            }
-            Some(n) => n,
-        };
+    fn result_rows(&mut self) -> () {
+        self.nrows = get_total_rows(&mut self.conn, &self.query)?;
     }
 
     #[throws(PostgresSourceError)]
@@ -264,7 +261,7 @@ where
         let pg_schema: Vec<_> = self.schema.iter().map(|&dt| dt.into()).collect();
         let iter = BinaryCopyOutIter::new(reader, &pg_schema);
 
-        PostgresBinarySourcePartitionParser::new(iter, &self.schema, self.buf_size)
+        PostgresBinarySourcePartitionParser::new(iter, &self.schema)
     }
 
     fn nrows(&self) -> usize {
@@ -288,12 +285,8 @@ where
     type Error = PostgresSourceError;
 
     #[throws(PostgresSourceError)]
-    fn prepare(&mut self) {
-        let row = self.conn.query_one(
-            count_query(&self.query, &PostgreSqlDialect {})?.as_str(),
-            &[],
-        )?;
-        self.nrows = row.get::<_, i64>(0) as usize;
+    fn result_rows(&mut self) {
+        self.nrows = get_total_rows(&mut self.conn, &self.query)?;
     }
 
     #[throws(PostgresSourceError)]
@@ -305,7 +298,7 @@ where
             .from_reader(reader)
             .into_records();
 
-        PostgresCSVSourceParser::new(iter, &self.schema, self.buf_size)
+        PostgresCSVSourceParser::new(iter, &self.schema)
     }
 
     fn nrows(&self) -> usize {
@@ -329,12 +322,8 @@ where
     type Error = PostgresSourceError;
 
     #[throws(PostgresSourceError)]
-    fn prepare(&mut self) {
-        let row = self.conn.query_one(
-            count_query(&self.query, &PostgreSqlDialect {})?.as_str(),
-            &[],
-        )?;
-        self.nrows = row.get::<_, i64>(0) as usize;
+    fn result_rows(&mut self) {
+        self.nrows = get_total_rows(&mut self.conn, &self.query)?;
     }
 
     #[throws(PostgresSourceError)]
@@ -342,7 +331,7 @@ where
         let iter = self
             .conn
             .query_raw::<_, bool, _>(self.query.as_str(), vec![])?; // unless reading the data, it seems like issue the query is fast
-        PostgresRawSourceParser::new(iter, &self.schema, self.buf_size)
+        PostgresRawSourceParser::new(iter, &self.schema)
     }
 
     fn nrows(&self) -> usize {
@@ -355,7 +344,6 @@ where
 }
 pub struct PostgresBinarySourcePartitionParser<'a> {
     iter: BinaryCopyOutIter<'a>,
-    buf_size: usize,
     rowbuf: Vec<BinaryCopyOutRow>,
     ncols: usize,
     current_col: usize,
@@ -363,15 +351,10 @@ pub struct PostgresBinarySourcePartitionParser<'a> {
 }
 
 impl<'a> PostgresBinarySourcePartitionParser<'a> {
-    pub fn new(
-        iter: BinaryCopyOutIter<'a>,
-        schema: &[PostgresTypeSystem],
-        buf_size: usize,
-    ) -> Self {
+    pub fn new(iter: BinaryCopyOutIter<'a>, schema: &[PostgresTypeSystem]) -> Self {
         Self {
             iter,
-            buf_size,
-            rowbuf: Vec::with_capacity(buf_size),
+            rowbuf: Vec::with_capacity(DB_BUFFER_SIZE),
             ncols: schema.len(),
             current_row: 0,
             current_col: 0,
@@ -380,27 +363,6 @@ impl<'a> PostgresBinarySourcePartitionParser<'a> {
 
     #[throws(PostgresSourceError)]
     fn next_loc(&mut self) -> (usize, usize) {
-        if self.current_row >= self.rowbuf.len() {
-            if !self.rowbuf.is_empty() {
-                self.rowbuf.drain(..);
-            }
-
-            for _ in 0..self.buf_size {
-                match self.iter.next()? {
-                    Some(row) => {
-                        self.rowbuf.push(row);
-                    }
-                    None => break,
-                }
-            }
-
-            if self.rowbuf.is_empty() {
-                throw!(anyhow!("Postgres EOF"));
-            }
-            self.current_row = 0;
-            self.current_col = 0;
-        }
-
         let ret = (self.current_row, self.current_col);
         self.current_row += (self.current_col + 1) / self.ncols;
         self.current_col = (self.current_col + 1) % self.ncols;
@@ -411,6 +373,24 @@ impl<'a> PostgresBinarySourcePartitionParser<'a> {
 impl<'a> PartitionParser<'a> for PostgresBinarySourcePartitionParser<'a> {
     type TypeSystem = PostgresTypeSystem;
     type Error = PostgresSourceError;
+
+    #[throws(PostgresSourceError)]
+    fn fetch_next(&mut self) -> (usize, bool) {
+        if !self.rowbuf.is_empty() {
+            self.rowbuf.drain(..);
+        }
+        for _ in 0..DB_BUFFER_SIZE {
+            match self.iter.next()? {
+                Some(row) => {
+                    self.rowbuf.push(row);
+                }
+                None => break,
+            }
+        }
+        self.current_row = 0;
+        self.current_col = 0;
+        (self.rowbuf.len(), self.rowbuf.len() < DB_BUFFER_SIZE)
+    }
 }
 
 macro_rules! impl_produce {
@@ -451,6 +431,12 @@ impl_produce!(
     f32,
     f64,
     Decimal,
+    Vec<i16>,
+    Vec<i32>,
+    Vec<i64>,
+    Vec<f32>,
+    Vec<f64>,
+    Vec<Decimal>,
     bool,
     &'r str,
     Vec<u8>,
@@ -464,7 +450,6 @@ impl_produce!(
 
 pub struct PostgresCSVSourceParser<'a> {
     iter: StringRecordsIntoIter<CopyOutReader<'a>>,
-    buf_size: usize,
     rowbuf: Vec<StringRecord>,
     ncols: usize,
     current_col: usize,
@@ -475,12 +460,10 @@ impl<'a> PostgresCSVSourceParser<'a> {
     pub fn new(
         iter: StringRecordsIntoIter<CopyOutReader<'a>>,
         schema: &[PostgresTypeSystem],
-        buf_size: usize,
     ) -> Self {
         Self {
             iter,
-            buf_size,
-            rowbuf: Vec::with_capacity(buf_size),
+            rowbuf: Vec::with_capacity(DB_BUFFER_SIZE),
             ncols: schema.len(),
             current_row: 0,
             current_col: 0,
@@ -489,26 +472,6 @@ impl<'a> PostgresCSVSourceParser<'a> {
 
     #[throws(PostgresSourceError)]
     fn next_loc(&mut self) -> (usize, usize) {
-        if self.current_row >= self.rowbuf.len() {
-            if !self.rowbuf.is_empty() {
-                self.rowbuf.drain(..);
-            }
-
-            for _ in 0..self.buf_size {
-                if let Some(row) = self.iter.next() {
-                    self.rowbuf.push(row?);
-                } else {
-                    break;
-                }
-            }
-
-            if self.rowbuf.is_empty() {
-                throw!(anyhow!("Postgres EOF"));
-            }
-            self.current_row = 0;
-            self.current_col = 0;
-        }
-
         let ret = (self.current_row, self.current_col);
         self.current_row += (self.current_col + 1) / self.ncols;
         self.current_col = (self.current_col + 1) % self.ncols;
@@ -519,6 +482,23 @@ impl<'a> PostgresCSVSourceParser<'a> {
 impl<'a> PartitionParser<'a> for PostgresCSVSourceParser<'a> {
     type Error = PostgresSourceError;
     type TypeSystem = PostgresTypeSystem;
+
+    #[throws(PostgresSourceError)]
+    fn fetch_next(&mut self) -> (usize, bool) {
+        if !self.rowbuf.is_empty() {
+            self.rowbuf.drain(..);
+        }
+        for _ in 0..DB_BUFFER_SIZE {
+            if let Some(row) = self.iter.next() {
+                self.rowbuf.push(row?);
+            } else {
+                break;
+            }
+        }
+        self.current_row = 0;
+        self.current_col = 0;
+        (self.rowbuf.len(), self.rowbuf.len() < DB_BUFFER_SIZE)
+    }
 }
 
 macro_rules! impl_csv_produce {
@@ -556,6 +536,59 @@ macro_rules! impl_csv_produce {
 
 impl_csv_produce!(i8, i16, i32, i64, f32, f64, Decimal, Uuid,);
 
+macro_rules! impl_csv_vec_produce {
+    ($($t: ty,)+) => {
+        $(
+            impl<'r, 'a> Produce<'r, Vec<$t>> for PostgresCSVSourceParser<'a> {
+                type Error = PostgresSourceError;
+
+                #[throws(PostgresSourceError)]
+                fn produce(&mut self) -> Vec<$t> {
+                    let (ridx, cidx) = self.next_loc()?;
+                    let s = &self.rowbuf[ridx][cidx][..];
+                    match s {
+                        "{}" => vec![],
+                        _ if s.len() < 3 => throw!(ConnectorXError::cannot_produce::<$t>(Some(s.into()))),
+                        s => s[1..s.len() - 1]
+                            .split(",")
+                            .map(|v| {
+                                v.parse()
+                                    .map_err(|_| ConnectorXError::cannot_produce::<$t>(Some(s.into())))
+                            })
+                            .collect::<Result<Vec<$t>, ConnectorXError>>()?,
+                    }
+                }
+            }
+
+            impl<'r, 'a> Produce<'r, Option<Vec<$t>>> for PostgresCSVSourceParser<'a> {
+                type Error = PostgresSourceError;
+
+                #[throws(PostgresSourceError)]
+                fn produce(&mut self) -> Option<Vec<$t>> {
+                    let (ridx, cidx) = self.next_loc()?;
+                    let s = &self.rowbuf[ridx][cidx][..];
+                    match s {
+                        "" => None,
+                        "{}" => Some(vec![]),
+                        _ if s.len() < 3 => throw!(ConnectorXError::cannot_produce::<$t>(Some(s.into()))),
+                        s => Some(
+                            s[1..s.len() - 1]
+                                .split(",")
+                                .map(|v| {
+                                    v.parse()
+                                        .map_err(|_| ConnectorXError::cannot_produce::<$t>(Some(s.into())))
+                                })
+                                .collect::<Result<Vec<$t>, ConnectorXError>>()?,
+                        ),
+                    }
+                }
+            }
+        )+
+    };
+}
+
+impl_csv_vec_produce!(i8, i16, i32, i64, f32, f64, Decimal,);
+
 impl<'r, 'a> Produce<'r, bool> for PostgresCSVSourceParser<'a> {
     type Error = PostgresSourceError;
 
@@ -597,7 +630,9 @@ impl<'r, 'a> Produce<'r, DateTime<Utc>> for PostgresCSVSourceParser<'a> {
     #[throws(PostgresSourceError)]
     fn produce(&mut self) -> DateTime<Utc> {
         let (ridx, cidx) = self.next_loc()?;
-        self.rowbuf[ridx][cidx].parse().map_err(|_| {
+        let s: &str = &self.rowbuf[ridx][cidx][..];
+        // postgres csv return example: 1970-01-01 00:00:01+00
+        format!("{}:00", s).parse().map_err(|_| {
             ConnectorXError::cannot_produce::<DateTime<Utc>>(Some(self.rowbuf[ridx][cidx].into()))
         })?
     }
@@ -612,7 +647,8 @@ impl<'r, 'a> Produce<'r, Option<DateTime<Utc>>> for PostgresCSVSourceParser<'a>
         match &self.rowbuf[ridx][cidx][..] {
             "" => None,
             v => {
-                Some(v.parse().map_err(|_| {
+                // postgres csv return example: 1970-01-01 00:00:01+00
+                Some(format!("{}:00", v).parse().map_err(|_| {
                     ConnectorXError::cannot_produce::<DateTime<Utc>>(Some(v.into()))
                 })?)
             }
@@ -785,7 +821,6 @@ impl<'r, 'a> Produce<'r, Option<Value>> for PostgresCSVSourceParser<'a> {
 
 pub struct PostgresRawSourceParser<'a> {
     iter: RowIter<'a>,
-    buf_size: usize,
     rowbuf: Vec<Row>,
     ncols: usize,
     current_col: usize,
@@ -793,11 +828,10 @@ pub struct PostgresRawSourceParser<'a> {
 }
 
 impl<'a> PostgresRawSourceParser<'a> {
-    pub fn new(iter: RowIter<'a>, schema: &[PostgresTypeSystem], buf_size: usize) -> Self {
+    pub fn new(iter: RowIter<'a>, schema: &[PostgresTypeSystem]) -> Self {
         Self {
             iter,
-            buf_size,
-            rowbuf: Vec::with_capacity(buf_size),
+            rowbuf: Vec::with_capacity(DB_BUFFER_SIZE),
             ncols: schema.len(),
             current_row: 0,
             current_col: 0,
@@ -806,26 +840,6 @@ impl<'a> PostgresRawSourceParser<'a> {
 
     #[throws(PostgresSourceError)]
     fn next_loc(&mut self) -> (usize, usize) {
-        if self.current_row >= self.rowbuf.len() {
-            if !self.rowbuf.is_empty() {
-                self.rowbuf.drain(..);
-            }
-
-            for _ in 0..self.buf_size {
-                if let Some(row) = self.iter.next()? {
-                    self.rowbuf.push(row);
-                } else {
-                    break;
-                }
-            }
-
-            if self.rowbuf.is_empty() {
-                throw!(anyhow!("Postgres EOF"));
-            }
-            self.current_row = 0;
-            self.current_col = 0;
-        }
-
         let ret = (self.current_row, self.current_col);
         self.current_row += (self.current_col + 1) / self.ncols;
         self.current_col = (self.current_col + 1) % self.ncols;
@@ -836,6 +850,23 @@ impl<'a> PostgresRawSourceParser<'a> {
 impl<'a> PartitionParser<'a> for PostgresRawSourceParser<'a> {
     type TypeSystem = PostgresTypeSystem;
     type Error = PostgresSourceError;
+
+    #[throws(PostgresSourceError)]
+    fn fetch_next(&mut self) -> (usize, bool) {
+        if !self.rowbuf.is_empty() {
+            self.rowbuf.drain(..);
+        }
+        for _ in 0..DB_BUFFER_SIZE {
+            if let Some(row) = self.iter.next()? {
+                self.rowbuf.push(row);
+            } else {
+                break;
+            }
+        }
+        self.current_row = 0;
+        self.current_col = 0;
+        (self.rowbuf.len(), self.rowbuf.len() < DB_BUFFER_SIZE)
+    }
 }
 
 macro_rules! impl_produce {
@@ -876,6 +907,12 @@ impl_produce!(
     f32,
     f64,
     Decimal,
+    Vec<i16>,
+    Vec<i32>,
+    Vec<i64>,
+    Vec<f32>,
+    Vec<f64>,
+    Vec<Decimal>,
     bool,
     &'r str,
     Vec<u8>,
